#Author-Team 07
#Description-Converts selected shape into lego bricks for fast prototyping and enables a brick-3Dprint combination.

import adsk.core, adsk.fusion, adsk.cam, traceback
import math, random

_ui = None
_app = None
_design = None

_handlers = []
_originalModel = None
_brickedBody = None

PLATE_HEIGHT = 0.32
BRICK_LENGHT_UNIT = 0.8
PIN_DISTANCE = 0.39
PIN_HEIGHT = 0.17
PIN_RADIUS = 0.24
BRICK_WALL_THICKNESS = 0.12
PIN_VOLUME = 0.03076247527


def run(context):
    try:
        global _app, _ui
        _app = adsk.core.Application.get()
        _ui = _app.userInterface

        # Get the UserInterface object and the CommandDefinitions collection.
        cmdDefs = _ui.commandDefinitions

        # Create a button command definition.
        brickifyButton = cmdDefs.addButtonDefinition('brickifyButton', 'Brickify', 'Converts selected shape into toy bricks and combines it with 3D-print models', 'icons/brick_2x4')

        # Connect to the command created event.
        brickifyButtonCreated = BrickifyButtonCreatedEventHandler()
        brickifyButton.commandCreated.add(brickifyButtonCreated)
        _handlers.append(brickifyButtonCreated)

        onBrickifyStart = BrickifyStartEventHandler()
        _ui.commandStarting.add(onBrickifyStart)
        _handlers.append(onBrickifyStart)

        # Create brickify remover command definition
        brickRemover = cmdDefs.addButtonDefinition('BrickRemover', 'Remove Bricks for 3D Model', 'Removes Bricks where 3D prints will be placed')
        removalCommandCreated = BrickifyRemovalCreatedEventHandler()
        brickRemover.commandCreated.add(removalCommandCreated)
        _handlers.append(removalCommandCreated)

        # Get the ADD-INS panel in the model workspace.
        addInsPanel = _ui.allToolbarPanels.itemById('SolidScriptsAddinsPanel')

        # Add the button to the bottom.
        buttonControl = addInsPanel.controls.addCommand(brickifyButton)

        # Make the button available in the panel.
        buttonControl.isPromotedByDefault = True
        buttonControl.isPromoted = True
        
    except:
        if _ui:
            _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))

# Event handler for the commandStarting event.
class BrickifyStartEventHandler(adsk.core.ApplicationCommandEventHandler):
    def __init__(self):
        super().__init__()
    def notify(self, args):
        global _design
        eventArgs = adsk.core.ApplicationCommandEventArgs.cast(args)
        if eventArgs.commandId == 'brickifyButton':
            # make sure DDT is active every time the Brickify button is pressed
            _design = adsk.fusion.Design.cast(_app.activeProduct)
            _design.designType = adsk.fusion.DesignTypes.DirectDesignType
        


# Event handler for the commandCreated event.
class BrickifyButtonCreatedEventHandler(adsk.core.CommandCreatedEventHandler):
    """Create ui window that appears when starting brickify."""
    def __init__(self):
        super().__init__()

    def notify(self, args):
        eventArgs = adsk.core.CommandCreatedEventArgs.cast(args)
        cmd = eventArgs.command

        # Connect to the execute event.
        onExecute = BrickifyButtonCommandExecuteHandler()
        cmd.execute.add(onExecute)
        _handlers.append(onExecute)
        cmd = eventArgs.command
        # Get the CommandInputs collection to create new command inputs
        inputs = cmd.commandInputs

        # Selection Input
        selection = inputs.addSelectionInput('brickify_selection', 'Model', 'Select the 3D model which you want to brickify')
        selection.addSelectionFilter('SolidBodies')

        # Buttons to choose a brick size
        brickButtons = inputs.addButtonRowCommandInput('brickify_bricksize', 'Brick Size', False)
        brickButtons.listItems.add('Plate 1×1', True, 'icons/plate_1x1')
        brickButtons.listItems.add('Plate 2×2', False, 'icons/plate_2x2')
        brickButtons.listItems.add('Brick 1×1', False, 'icons/brick_1x1')
        brickButtons.listItems.add('Brick 2×2', False, 'icons/brick_2x2')
        
        # Buttons to choose the color
        brickColor = inputs.addDropDownCommandInput('brickifyColor', 'Choose color', adsk.core.DropDownStyles.LabeledIconDropDownStyle)
        brickColor.listItems.add('Colorful', True, 'icons/color/colorful')
        brickColor.listItems.add('Red', False, 'icons/color/red')
        brickColor.listItems.add('Green', False, 'icons/color/green')
        brickColor.listItems.add('Blue', False, 'icons/color/blue')
        brickColor.listItems.add('Yellow', False, 'icons/color/yellow')
        brickColor.listItems.add('White', False, 'icons/color/white')
        brickColor.listItems.add('Black', False, 'icons/color/black')
        brickColor.listItems.add('No Color', False, 'icons/color/none')



# Event handler for the execute event.
class BrickifyButtonCommandExecuteHandler(adsk.core.CommandEventHandler):
    """Is executed when the OK button was pressed in the main brickify window."""
    def __init__(self):
        super().__init__()

    def notify(self, args):
        eventArgs = adsk.core.CommandEventArgs.cast(args)
        inputs = eventArgs.command.commandInputs
        global _originalModel, _brickedBody

        # get values from the buttons
        colorMode = inputs.itemById('brickifyColor').selectedItem.index
        _originalModel = inputs.itemById('brickify_selection').selection(0).entity
        _originalModel.isSelectable = False 
        brickSize_input = inputs.itemById('brickify_bricksize')
        brickSize = get_brick_size(brickSize_input.selectedItem)

        # Brickify the model
        _brickedBody = brickify(_originalModel, brickSize, colorMode)

        # execute BrickifyRemoval handler
        _ui.commandDefinitions.itemById('BrickRemover').execute()



class BrickifyRemovalCreatedEventHandler(adsk.core.CommandCreatedEventHandler):
    """Creates the brick removal ui and connects BrickifyRemoval-handlers to events"""
    def __init__(self):
        super().__init__()

    def notify(self, args):
        eventArgs = adsk.core.CommandCreatedEventArgs.cast(args)
        cmd = eventArgs.command

        # Connect to the execute event.
        onExecute = BrickifyRemovalCommandExecuteHandler()
        cmd.execute.add(onExecute)
        _handlers.append(onExecute)
        # Connect to preview event
        onExecutePreview = BrickifyRemovalCommandExecutePreviewHandler()
        cmd.executePreview.add(onExecutePreview)
        _handlers.append(onExecutePreview)
        # Connect to input change event
        onChanged = BrickifyRemovalChangedHandler()
        cmd.inputChanged.add(onChanged)
        _handlers.append(onChanged)
       # Get the CommandInputs collection to create new command inputs
        inputs = cmd.commandInputs
        # Add a selection without maximum to the dialoge
        bodiesInput = inputs.addSelectionInput('removalSelection', 'Bricks', 'Select the bricks you want to exchange for an 3D model')
        bodiesInput.addSelectionFilter('SolidBodies')
        bodiesInput.setSelectionLimits(0)
        # Add checkbox for live removal
        boolSelectionInput = inputs.addBoolValueInput('boolSelection', 'Live removal', True, '', False)
        boolSelectionInput.isFullWidth
        boolHideOriginal = inputs.addBoolValueInput('boolHideOriginal', 'Hide original', True, '', False)
        boolHideOriginal.isFullWidth
        boolHideOriginal.isEnabled = False
        # Add a slider for opacity adjustment of bricks
        valueList = [10, 20, 30, 40, 50, 60, 70, 80, 90, 100]
        opacityInput = inputs.addIntegerSliderListCommandInput('opacitySlider', 'Brick opacity', valueList)
        opacityInput.setText("0%","100%")
        opacityInput.isFullWidth
        opacityInput.valueOne = 100


class BrickifyRemovalChangedHandler(adsk.core.InputChangedEventHandler):
    """Toggels the activation state of the 'boolHideOriginal' checkbox"""
    def __init__(self):
        super().__init__()

    def notify(self, args):
        try:
            eventArgs = adsk.core.InputChangedEventArgs.cast(args)
            inputs = eventArgs.inputs
            cmdInput = eventArgs.input
            # onInputChange for slider controller
            if cmdInput.id == "boolSelection":
                if inputs.itemById('boolSelection').value:
                    inputs.itemById('boolHideOriginal').isEnabled = True
                else: 
                    inputs.itemById('boolHideOriginal').isEnabled = False
        except:
            if _ui:
                _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))

class BrickifyRemovalCommandExecutePreviewHandler(adsk.core.CommandEventHandler):
    """Enables live brick opacity changes and live brick removal preview"""
    def __init__(self):
        super().__init__()

    def notify(self, args):
        try:
            eventArgs = adsk.core.CommandEventArgs.cast(args)
            inputs = eventArgs.command.commandInputs
            opacity = inputs.itemById('opacitySlider').valueOne / 100.0
            adjustOpacity(opacity)
            boolSelectionInput = inputs.itemById('boolSelection').value
            boolHideOriginal = inputs.itemById('boolHideOriginal').value
            selection = inputs.itemById('removalSelection')
            if boolSelectionInput:
                for index in range(0,selection.selectionCount):
                    body = adsk.fusion.BRepBody.cast(inputs.itemById('removalSelection').selection(index).entity)
                    body.isLightBulbOn = False
            if boolHideOriginal:
                _originalModel.isLightBulbOn = False
            else:
                _originalModel.isLightBulbOn = True
        except:
            if _ui:
                _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))
                

# Event handler for the execute event.
class BrickifyRemovalCommandExecuteHandler(adsk.core.CommandEventHandler):
    """Executes the removal of selected bricks and cuts out the print model"""
    def __init__(self):
        super().__init__()

    def notify(self, args):
        global _changedOpacity
        eventArgs = adsk.core.CommandEventArgs.cast(args)
        inputs = eventArgs.command.commandInputs
        selectedBricks = inputs.itemById('removalSelection') 
        removeBricks(selectedBricks)
        cutPrintModel(_brickedBody, _originalModel)
        #TODO open print model in new document            


def stop(context):
    try:
        # Clean up the _ui.
        cmdDef = _ui.commandDefinitions.itemById('brickifyButton')
        if cmdDef:
            cmdDef.deleteMe()
        cmdDef = _ui.commandDefinitions.itemById('removalSelection')
        if cmdDef:
            cmdDef.deleteMe()
        cmdDef = _ui.commandDefinitions.itemById('BrickRemover')
        if cmdDef:
            cmdDef.deleteMe()
        addinsPanel = _ui.allToolbarPanels.itemById('SolidScriptsAddinsPanel')
        cntrl = addinsPanel.controls.itemById('brickifyButton')
        if cntrl:
            cntrl.deleteMe()
        
    except:
        if _ui:
            _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))


def brickify(selectedObject, brickSize=(2, 2, 3), colorMode = 0):
    """Fill bounding box with bricks and delete those that do not touch the model."""
    try:
        if not _design:
            _ui.messageBox('No active Fusion 360 design', 'No Design')
            return
        size = getBoundingBoxDimensions(selectedObject)
        pos = getMinPoint(selectedObject)
        brickOccurrence = buildBrick(brickSize=brickSize)
        boundingBoxBlock = fillBoundingBox(brickOccurrence, brickSize=brickSize, pos=pos, size=size)
        trimmedBrickBlock = trimBrickBlock(boundingBoxBlock, selectedObject, colorMode)
        return trimmedBrickBlock

    except:
        if _ui:
            _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))

def get_brick_size(selectedBrick):
    """Map the button values to brick sizes."""
    name = selectedBrick.name
    if 'Brick' in name:
        height = 3
    else:
        height = 1

    if '1×1' in name:
        return (1, 1, height)
    if '1×2' in name:
        return (1, 2, height)
    if '1×4' in name:
        return (1, 4, height)
    if '2×2' in name:
        return (2, 2, height)
    if '2×4' in name:
        return (2, 4, height)
    return (2, 4, 3)


# build brick returns an occurrence of a brick to be reused in the model
def buildBrick(x=0, y=0, z=0, brickSize=(2, 2, 3)):
    """Build a brick of selected size at selected coordinates"""
    brickColumns = brickSize[0]
    brickRows = brickSize[1]
    brickHeight = brickSize[2] * PLATE_HEIGHT
    try:
        rootComp = _design.rootComponent

        #create an occurence and edit the component
        newOccurence = adsk.core.Matrix3D.create()
        brickOccurrence = rootComp.occurrences.addNewComponent(newOccurence)
        brickComponent = brickOccurrence.component
        #setting an appropriate name based on brick form
        brickComponent.name = 'Brick ' + str(brickColumns) + 'x' + str(brickRows)

        sketches = brickComponent.sketches
        xyPlane = brickComponent.xYConstructionPlane
        sketch = sketches.add(xyPlane)
        lines = sketch.sketchCurves.sketchLines

        # build base sketch
        point0 = adsk.core.Point3D.create(x, y, z)
        point1 = adsk.core.Point3D.create(x + (BRICK_LENGHT_UNIT * brickRows), y + (BRICK_LENGHT_UNIT * brickColumns), z)
        lines.addTwoPointRectangle(point0, point1)

        # extrude base to block
        brickShape = sketch.profiles.item(0)
        extruder = brickComponent.features.extrudeFeatures
        brickExtrusion = extruder.createInput(brickShape, adsk.fusion.FeatureOperations.NewBodyFeatureOperation)
        brickExtrusionHeight = adsk.core.ValueInput.createByReal(brickHeight)
        brickExtrusion.setDistanceExtent(False, brickExtrusionHeight)
        brickExtrusion.isSolid = True
        ex = extruder.add(brickExtrusion)

        # select brick body
        body = ex.bodies.item(0)
        # select brick body top face and add new sketch plane
        topFace = body.faces.item(4)
        bottomFace = body.faces.item(5)
        pinSketch = sketches.add(topFace)
        # add pin sketches and extrude,
        profileCollection = adsk.core.ObjectCollection.create()
        for i in range(brickColumns):
            for j in range(brickRows):
                tempSketch = pinSketch.sketchCurves.sketchCircles.addByCenterRadius(adsk.core.Point3D.create(
                                                                                        x + PIN_DISTANCE + j * BRICK_LENGHT_UNIT,
                                                                                        y + PIN_DISTANCE + i * BRICK_LENGHT_UNIT,
                                                                                        z),
                                                                                    PIN_RADIUS)
        for pinProfileID in range(1, brickRows * brickColumns + 1):
            profileCollection.add(pinSketch.profiles.item(pinProfileID))
        pinExtrusion = extruder.createInput(profileCollection, adsk.fusion.FeatureOperations.JoinFeatureOperation)
        pinExtrusionHeight = adsk.core.ValueInput.createByReal(PIN_HEIGHT)
        pinExtrusion.setDistanceExtent(False, pinExtrusionHeight)
        extruder.add(pinExtrusion)

        # cut into the brick by creating
        pinSketchBottom = sketches.add(bottomFace)
        profileCollectionBottom = adsk.core.ObjectCollection.create()
        for i in range(brickColumns):
            for j in range(brickRows):
                tempSketch = pinSketchBottom.sketchCurves.sketchCircles.addByCenterRadius(adsk.core.Point3D.create(
                                                                                            - (x + PIN_DISTANCE + j * BRICK_LENGHT_UNIT),
                                                                                            y + PIN_DISTANCE + i * BRICK_LENGHT_UNIT,
                                                                                            z),
                                                                                          PIN_RADIUS)
        for pinProfileID in range(1, brickRows * brickColumns + 1):
            profileCollectionBottom.add(pinSketchBottom.profiles.item(pinProfileID))
        pinBottomExtrusion = extruder.createInput(profileCollectionBottom, adsk.fusion.FeatureOperations.CutFeatureOperation)
        pinExtrusionHeightBottom = adsk.core.ValueInput.createByReal(-PIN_HEIGHT)
        pinBottomExtrusion.setDistanceExtent(False, pinExtrusionHeightBottom)
        extruder.add(pinBottomExtrusion)

        # remove sketches that are no longer useful
        sketch.deleteMe()
        pinSketch.deleteMe()
        pinSketchBottom.deleteMe()
        brickOccurrence.isLightBulbOn = False

        return brickOccurrence

    except:
        if _ui:
            _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))


def createBlock(brickOccurrence, brickSize=(2, 2, 3), sizeX=2, sizeY=2, sizeZ=3, pos=(0, 0, 0)):
    """Copy/paste a brick a set number of times to create a solid box."""
    try:
        rootComp = _design.rootComponent
        tempBRep = adsk.fusion.TemporaryBRepManager.get()
        transform = adsk.core.Matrix3D.create()
        bodyCollection = adsk.core.ObjectCollection.create()

        brickSizeX = brickSize[1]
        brickSizeY = brickSize[0]
        brickSizeZ = brickSize[2]
        # Adding progress bar
        # Extra whitepspace inserted because progressbars are renderered too short on macOS
        progressBar = addProgressBar('    Creating Bricks    ', '    Adding Brick %v of %m (%p %)   ', 0, sizeX*sizeY*sizeZ)
        brickBody = brickOccurrence.bRepBodies.item(0)
        boundingBoxBlock = rootComp.occurrences.addNewComponent(adsk.core.Matrix3D.create())
        for x in range(sizeX):
            for y in range(sizeY):
                for z in range(sizeZ):
                    # Calculate vector for translation
                    vector = adsk.core.Vector3D.create(
                        (pos[0] + x * brickSizeX * BRICK_LENGHT_UNIT),
                        (pos[1] + y * brickSizeY * BRICK_LENGHT_UNIT),
                        (pos[2] + z * brickSizeZ * PLATE_HEIGHT)
                    )
                    if vector.x == 0 and vector.y == 0 and vector.z == 0:
                        continue
                    transform.translation = vector
                    #use temporaryBRepManager to quickly copy and translate body from brickOccurence
                    bodyCopy = tempBRep.copy(brickBody)
                    tempBRep.transform(bodyCopy, transform)
                    boundingBoxBlock.component.bRepBodies.add(bodyCopy)
                    progressBar.progressValue = progressBar.progressValue+1
        hideProgressBar(progressBar)
        return boundingBoxBlock
    except:
        hideProgressBar(progressBar)
        if _ui:
            _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))


def getBoundingBoxDimensions(obj):
    """Calculate the bounding box dimensions."""
    bb = obj.boundingBox
    minPoint = bb.minPoint
    maxPoint = bb.maxPoint

    sizeX = maxPoint.x - minPoint.x
    sizeY = maxPoint.y - minPoint.y
    sizeZ = maxPoint.z - minPoint.z

    return (sizeX, sizeY, sizeZ)


def getMinPoint(obj):
    """Get point with the smallest coordinates in the body"""
    bb = obj.boundingBox
    minPoint = bb.minPoint

    return (minPoint.x, minPoint.y, minPoint.z)

def fillBoundingBox(brickOccurrence, brickSize=(4, 2, 3), pos=(0, 0, 0), size=(2, 2, 2)):
    """Fill a bounding box with the brick occurence. (box size in cm, brick dimensions in columns and rows)"""
    try:
        sizeX = size[0]
        sizeY = size[1]
        sizeZ = size[2]

        brickSizeX = brickSize[1] * BRICK_LENGHT_UNIT
        brickSizeY = brickSize[0] * BRICK_LENGHT_UNIT
        brickSizeZ = brickSize[2] * PLATE_HEIGHT

        brickCountX = math.ceil(sizeX/brickSizeX)
        brickCountY = math.ceil(sizeY/brickSizeY)
        brickCountZ = math.ceil(sizeZ/brickSizeZ)

        boundingBoxBlock = createBlock(brickOccurrence, brickSize, brickCountX, brickCountY, brickCountZ, pos)
        boundingBoxBlock.component.name = f'{_originalModel.name} (Brickified)'
        return boundingBoxBlock
    except:
        if _ui:
            _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))


def trimBrickBlock(boundingBoxBlock, selectedObject, colorMode):
    """Trim the bounding box block of lego bricks to fit the shape of the original object."""
    try:
        inputOccurrences = adsk.core.ObjectCollection.create()
        inputOccurrences.add(boundingBoxBlock)
        inputOccurrences.add(selectedObject)

        progressBar = addProgressBar(' Working out model details    ', '  Calculating interference (%p %)        ', 0, boundingBoxBlock.bRepBodies.count)
        # Create the interferenceInput object and run the analysis.
        interferenceInput = _design.createInterferenceInput(inputOccurrences)
        interferenceInput.areCoincidentFacesIncluded = False
        results = _design.analyzeInterference(interferenceInput)

        interferenceBodies = []
        unusedBricks = adsk.core.ObjectCollection.create()
        #note: syntax "for body in results" causes out of bounds exception here
        for index in range(0,results.count):
            interferenceBodies.append(results.item(index).entityOne.name)

        #note: syntax "for body in bB.bRepBodies" causes out of bounds exception here
        for i in range(0,boundingBoxBlock.bRepBodies.count):
            if boundingBoxBlock.bRepBodies.item(i):
                body = boundingBoxBlock.bRepBodies.item(i)
                body.appearance = getColor(colorMode)
                progressBar.progressValue = progressBar.progressValue+1
                #cannot delete unused objects right away, because the index is affected
                if body.name not in interferenceBodies:
                    unusedBricks.add(body)
        for brick in range(0,unusedBricks.count):
            unusedBricks.item(brick).deleteMe()
        hideProgressBar(progressBar)
        return boundingBoxBlock
    except:
        hideProgressBar(progressBar)
        if _ui:
            _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))


def removeBricks(brickSelection):
    """Removes all bricks in the argument except the originally selected body"""
    try:
        bricks = []
        # selections are fragile, so the selected blocks are added into the bricks array
        for i in range (0, brickSelection.selectionCount):
            bricks.append(brickSelection.selection(i).entity)
        for brick in bricks:
            brick.deleteMe()
    except:
        if _ui:
            _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))

def adjustOpacity(opacityValue):
    """Adjusts the opacity of the brickified Model"""
    try:
        bodies = _brickedBody.bRepBodies
        for index in range(0,bodies.count):
            bodies.item(index).opacity = opacityValue
    except:
        if _ui:
            _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))  

def getColor(colorMode = 0):
    """returns 'enamel glossy appearance object from  in accordance to the selected color mode"""
    fusionMaterials = _app.materialLibraries.itemByName("Fusion 360 Appearance Library").appearances
    colors= {
        1: "Paint - Enamel Glossy (Red)",
        2: "Paint - Enamel Glossy (Green)",
        3: "Paint - Enamel Glossy (Blue)",
        4: "Paint - Enamel Glossy (Yellow)",
        5: "Paint - Enamel Glossy (White)",
        6: "Paint - Enamel Glossy (Black)"
    }
    if colorMode == 0:
        colorMode = random.randint(1,6)
    if colorMode == 7:
        return None
    return fusionMaterials.itemByName(colors[colorMode])

def cutPrintModel(brickModel, originalModel):
    """Cuts extruding parts from the original Model"""
    try:
        rootComp = _design.rootComponent
        inputOccurrences = adsk.core.ObjectCollection.create()
        progressBar = addProgressBar('    Cut out model parts    ', '    Cutting out non-interfering parts and cleaning up dust (%p %)    ', 0, brickModel.bRepBodies.count + 1)
        # Add all bricks to a collection
        for index in range(0,brickModel.bRepBodies.count):
            brick = brickModel.bRepBodies.item(index)
            inputOccurrences.add(brick)
            progressBar.progressValue = progressBar.progressValue + 1
        combineCutFeats = rootComp.features.combineFeatures
        cutInterference = combineCutFeats.createInput(originalModel, inputOccurrences)
        cutInterference.operation = adsk.fusion.FeatureOperations.CutFeatureOperation
        cutInterference.isKeepToolBodies = True
        combineCutFeats.add(cutInterference)
        bodies = rootComp.bRepBodies
        printBodies = adsk.core.ObjectCollection.create()
        rmBodies = []
        for i in range(0,bodies.count):
            body = bodies.item(i)
            body.isSelectable = True
            if body.volume <= PIN_VOLUME:
                rmBodies.append(body)
            else:
                printBodies.add(body)
        # necessary workaround for progressDialog
        for x in rmBodies:
            x.deleteMe()
        hideProgressBar(progressBar)
        return printBodies
    except:
        hideProgressBar(progressBar)
        if _ui:
            _ui.messageBox('Failed:\n{}'.format(traceback.format_exc()))  

def addProgressBar(title, label, minValue, maxValue, delay = 0):
    """Adds non-tranclucent progress bar without cancel-button"""
    progressDialog = _ui.createProgressDialog()
    progressDialog.cancelButtonText = 'Cancel'
    progressDialog.isBackgroundTranslucent = False
    progressDialog.isCancelButtonShown = False
    progressDialog.show(title, label, minValue, maxValue, delay)
    adsk.doEvents()
    return progressDialog

def hideProgressBar(progressBarObject):
    """Hides ProgressBar and refreshes events and viewport to ensure hiding"""
    if(progressBarObject):
        progressBarObject.hide()
        adsk.doEvents()
        _app.activeViewport.refresh()
        